@@ -42,6 +42,8 @@ PATH="$PATH:/usr/sbin:/sbin"
 # libtest_exit_cmds+=(expr).
 libtest_exit_cmds=()
 run_exit_cmds() {
+  # Quiet termination
+  set +x
   for expr in "${libtest_exit_cmds[@]}"; do
     eval "${expr}" || true
   done
@@ -283,6 +285,26 @@ ostree_repo_init() {
     fi
 }
 
+run_webserver() {
+    echo httpd=${OSTREE_HTTPD}
+    if test -z "${OSTREE_HTTPD:-}"; then
+        if test "$#" -gt 0; then
+            echo "fatal: unhandled arguments for webserver: $@" 1>&2
+            exit 1
+        fi
+        # Note this automatically daemonizes; close stdin to ensure it doesn't leak to the child.
+        ${test_srcdir}/webserver.py ${test_tmpdir}/httpd-port </dev/null &>/dev/null
+        echo "Waiting for webserver..."
+        while test '!' -f ${test_tmpdir}/httpd-port; do
+            sleep 0.5
+        done
+    else
+        ${OSTREE_HTTPD} --autoexit --log-file $(pwd)/httpd.log --daemonize -p ${test_tmpdir}/httpd-port "$@"
+    fi
+    port=$(cat ${test_tmpdir}/httpd-port)
+    echo "http://127.0.0.1:${port}" > ${test_tmpdir}/httpd-address
+}
+
 # The original one; use setup_fake_remote_repo2 for newer code,
 # down the line we'll try to port tests.
 setup_fake_remote_repo1() {
@@ -314,9 +336,7 @@ setup_fake_remote_repo1() {
     mkdir ${test_tmpdir}/httpd
     cd httpd
     ln -s ${test_tmpdir}/ostree-srv ostree
-    ${OSTREE_HTTPD} --autoexit --log-file $(pwd)/httpd.log --daemonize -p ${test_tmpdir}/httpd-port "$@"
-    port=$(cat ${test_tmpdir}/httpd-port)
-    echo "http://127.0.0.1:${port}" > ${test_tmpdir}/httpd-address
+    run_webserver "$@"
     cd ${oldpwd} 
 
     export OSTREE="${CMD_PREFIX} ostree --repo=repo"
@@ -359,9 +379,7 @@ setup_fake_remote_repo2() {
     mkdir ${test_tmpdir}/httpd
     cd httpd
     ln -s ${test_tmpdir}/ostree-srv ostree
-    ${OSTREE_HTTPD} --autoexit --log-file $(pwd)/httpd.log --daemonize -p ${test_tmpdir}/httpd-port $args
-    port=$(cat ${test_tmpdir}/httpd-port)
-    echo "http://127.0.0.1:${port}" > ${test_tmpdir}/httpd-address
+    run_webserver $args
     cd ${oldpwd}
     export OSTREE="${CMD_PREFIX} ostree --repo=repo"
 }
@@ -525,9 +543,7 @@ EOF
     mkdir ${test_tmpdir}/httpd
     cd httpd
     ln -s ${test_tmpdir} ostree
-    ${OSTREE_HTTPD} --autoexit --daemonize -p ${test_tmpdir}/httpd-port
-    port=$(cat ${test_tmpdir}/httpd-port)
-    echo "http://127.0.0.1:${port}" > ${test_tmpdir}/httpd-address
+    run_webserver
     cd ${oldpwd} 
 }
 
@@ -612,6 +628,18 @@ skip_one_without_user_xattrs () {
     fi
 }
 
+skip_without_ostree_httpd () {
+    if test -z "${OSTREE_HTTPD:-}"; then
+        skip "this test requires libsoup (ostree-trivial-httpd)"
+    fi
+}
+
+skip_known_xfail_docker() {
+    if test "${OSTREE_TEST_SKIP:-}" = known-xfail-docker; then
+        skip "This test was explicitly skipped via OSTREE_TEST_SKIP=known-xfail-docker"
+    fi
+}
+
 skip_without_user_xattrs () {
     if ! have_user_xattrs; then
         skip "this test requires xattr support"
@@ -747,8 +775,9 @@ which_gpg () {
 }
 
 libtest_cleanup_gpg () {
+    set +x
     local gpg_homedir=${1:-${test_tmpdir}/gpghome}
-    gpg-connect-agent --homedir "${gpg_homedir}" killagent /bye || true
+    gpg-connect-agent --homedir "${gpg_homedir}" killagent /bye &>/dev/null || true
 }
 libtest_exit_cmds+=(libtest_cleanup_gpg)
 
